/********************************************************************
* Description: XZABkins.c
*   Derived from trivkins for 3 axis Cartesian machine
*   Established for 4-axis lathe (XZAB) where A rotates around X in Z-Yplane and B rotates around Y in Z-Xplane
*
*   Derived from a work by Fred Proctor & Will Shackleford
*
* License: GPL Version 2
*    
* Copyright (c) 2009 All rights reserved.
*
********************************************************************/

#include "motion.h"
#include "hal.h"
#include "rtapi.h"
#include "rtapi.h"		/* RTAPI realtime OS API */
#include "rtapi_app.h"		/* RTAPI realtime module decls */
#include "rtapi_math.h"
#include "rtapi_string.h"

struct haldata { 
    hal_s32_t joints[EMCMOT_MAX_JOINTS];
    hal_float_t *Tool_offset;
    hal_s32_t *Tool_number;
    hal_float_t *J0;
    hal_float_t *J2;
    hal_float_t *J4;

} *haldata;

int bAxis = 4;

hal_float_t Last_Tool_offset = 0.00;
hal_s32_t Last_Tool_number = 0;


int kinematicsForward(const double *joints,
		      EmcPose * pos,
		      const KINEMATICS_FORWARD_FLAGS * fflags,
		      KINEMATICS_INVERSE_FLAGS * iflags)
{
    double b_rad;
    double dT = *(haldata->Tool_offset);
    long lTN = *(haldata->Tool_number);
    
    if (lTN == 100) {
	b_rad = (joints[4]-90)*M_PI/180; /* B-axis offset */
    } else {
	b_rad = (joints[4])*M_PI/180; /* B-axis offset */
    }

    pos->tran.x = joints[0] - (dT * sin(b_rad));
    pos->tran.y = joints[1];
    pos->tran.z = joints[2] - (dT * cos(b_rad)) + dT; 
    pos->a = joints[3];
    pos->b = joints[4];
    pos->c = joints[5];
    pos->u = joints[6];
    pos->v = joints[7];
    pos->w = joints[8];

    return 0;
}

int kinematicsInverse(const EmcPose * pos,
		      double *joints,
		      const KINEMATICS_INVERSE_FLAGS * iflags,
		      KINEMATICS_FORWARD_FLAGS * fflags)
{
    double b_rad;
    double dT = *(haldata->Tool_offset);
    long lTN = *(haldata->Tool_number);

    if (lTN == 100) {
	b_rad = (pos->b-90)*M_PI/180; /* B-axis offset */
    } else {
	b_rad = (pos->b)*M_PI/180; /* B-axis offset */
    }

    joints[0] = pos->tran.x + (dT * sin(b_rad));
    joints[1] = pos->tran.y;
    joints[2] = (pos->tran.z - dT) +  (dT * cos(b_rad)) ;
    joints[3] = pos->a;
    joints[4] = pos->b;
    joints[5] = pos->c;
    joints[6] = pos->u;
    joints[7] = pos->v;
    joints[8] = pos->w;

    *(haldata->J0) = joints[0];
    *(haldata->J2) = joints[2];
    *(haldata->J4) = joints[4];

    return 0;
}

/* implemented for these kinematics as giving joints preference */
int kinematicsHome(EmcPose * world,
		   double *joint,
		   KINEMATICS_FORWARD_FLAGS * fflags,
		   KINEMATICS_INVERSE_FLAGS * iflags)
{
    *fflags = 0;
    *iflags = 0;

    return kinematicsForward(joint, world, fflags, iflags);
}

static KINEMATICS_TYPE ktype = -1;

KINEMATICS_TYPE kinematicsType()
{
    return ktype;
}

static char *coordinates = "XYZABCUVW";
RTAPI_MP_STRING(coordinates, "Existing Axes");

static char *kinstype = "1"; // use KINEMATICS_IDENTITY
RTAPI_MP_STRING(kinstype, "Kinematics Type (Identity,Both)");

EXPORT_SYMBOL(kinematicsType);
EXPORT_SYMBOL(kinematicsForward);
EXPORT_SYMBOL(kinematicsInverse);
MODULE_LICENSE("GPL");

static int next_axis_number(void) {
    while(*coordinates) {
	switch(*coordinates) {
	    case 'x': case 'X': coordinates++; return 0;
	    case 'y': case 'Y': coordinates++; return 1;
	    case 'z': case 'Z': coordinates++; return 2;
	    case 'a': case 'A': coordinates++; return 3;
	    case 'b': case 'B': coordinates++; return 4;
	    case 'c': case 'C': coordinates++; return 5;
	    case 'u': case 'U': coordinates++; return 6;
	    case 'v': case 'V': coordinates++; return 7;
	    case 'w': case 'W': coordinates++; return 8;
	    case ' ': case '\t': coordinates++; continue;
	}
	rtapi_print_msg(RTAPI_MSG_ERR,
		"XZABkins: ERROR: Invalid character '%c' in coordinates\n",
		*coordinates);
		return -1;
    }
    return -1;
}
int comp_id;
int rtapi_app_main(void) {
    int i;
    int res = 0;
    comp_id = hal_init("XZABkins");
    if(comp_id < 0) return comp_id;

    haldata = hal_malloc(sizeof(struct haldata));

    for(i=0; i<EMCMOT_MAX_JOINTS; i++) {
	haldata->joints[i] = next_axis_number();
    }
    switch (*kinstype) {
      case 'b': case 'B': ktype = KINEMATICS_BOTH;         break;
      case 'f': case 'F': ktype = KINEMATICS_FORWARD_ONLY; break;
      case 'i': case 'I': ktype = KINEMATICS_INVERSE_ONLY; break;
      case '1': default:  ktype = KINEMATICS_IDENTITY;
    }

    if((res = hal_pin_float_new("XZABkins.Tool-offset", HAL_IN, &(haldata->Tool_offset),
      comp_id)) < 0) goto error;
    if((res = hal_pin_s32_new("XZABkins.Tool-number", HAL_IN, &(haldata->Tool_number),
      comp_id)) < 0) goto error;
    if((res = hal_pin_float_new("XZABkins.Joint0", HAL_OUT, &(haldata->J0),
      comp_id)) < 0) goto error;
    if((res = hal_pin_float_new("XZABkins.Joint2", HAL_OUT, &(haldata->J2),
      comp_id)) < 0) goto error;
    if((res = hal_pin_float_new("XZABkins.Joint4", HAL_OUT, &(haldata->J4),
      comp_id)) < 0) goto error;

    hal_ready(comp_id);
    return 0;

error:
    hal_exit(comp_id);
    return res;
}

void rtapi_app_exit(void) { hal_exit(comp_id); }
